package com.lnb.multidatasourcedemo.datasource;

import com.alibaba.druid.pool.DruidDataSource;
import com.alibaba.druid.pool.DruidPooledConnection;
import com.lnb.multidatasourcedemo.constant.Constant;
import com.lnb.multidatasourcedemo.curator.ZkProperties;
import com.lnb.multidatasourcedemo.model.DatasourceModel;
import com.lnb.multidatasourcedemo.model.RoutingKey2DB;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

@Component
@Slf4j
public class DbAdapter {
    @Autowired
    private ZkProperties properties;

    private DruidDataSource dataSource;

    /* 配置jdbc url */
    public static final String SOCKET = "socketTimeout=";
    /* 配置jdbc socket timeout 60s */
    public static final String TIMEOUT = "60000";

    private Map<String, DbUrl> db_ID = new HashMap<>();

    /**
     * 初始化数据源
     * @param map
     */
    public void initDatasource(Map<String, DruidDataSource> map) {
        log.info("initDatasource start");
        DruidPooledConnection druidConn = null;
        Connection conn = null;
        PreparedStatement preparedStatement = null;
        ResultSet resultSet = null;

        int index = 1;

        //获取配置文件中的appName，然后查询对应的数据库链接信息
        Set<String> appNameSet = properties.getAppName();
        String query = "SELECT dbID, appName, address, port, dbName, username, password FROM tbl_sys_db";
        String condition = " where appName in (";
        String order = " ORDER BY dbID;";
        String queryCombine = createDbQuery(query, condition, order, appNameSet);
        try {
            druidConn = dataSource.getConnection();
            conn = druidConn.getConnection();
            preparedStatement = conn.prepareStatement(queryCombine);
            for (String appName : appNameSet) {
                preparedStatement.setString(index, appName);
                index++;
            }
            resultSet = preparedStatement.executeQuery();
            DbUrl dbUrl;
            while (resultSet.next()) {
                String dbID = resultSet.getString("dbID");
                String address = resultSet.getString("address");
                String port = resultSet.getString("port");
                String username = resultSet.getString("username");
                String password = resultSet.getString("password");
                String dbName = resultSet.getString("dbName");
                String url = address + ":" + port;
                DatasourceModel datasourceModel = new DatasourceModel();
                datasourceModel.setUrl(url);
                datasourceModel.setCommonDB("?useUnicode=true&characterEncoding=utf8&allowMultiQueries=true&rewriteBatchedStatements=true");
                datasourceModel.setUrlPrefix("jdbc:mysql://");
                datasourceModel.setUsername(username);
                datasourceModel.setPwd(password);
                putDataSource(map, datasourceModel, properties);

                //把appName和数据库连接地址的关系存起来，处理routingKeys的时候获取他两之间的关系
                dbUrl = new DbUrl();
                dbUrl.setDbName(dbName);
                dbUrl.setUrl(url);
                db_ID.put(dbID, dbUrl);
            }
            log.info("initDatasource finished successfully");
        } catch (SQLException e) {
            log.error("read tbl_sys_db hit error!", e);
        } finally {
            close(druidConn, conn, preparedStatement, resultSet);
        }
    }

    /**
     * 关闭数据库连接
     * @param druidConn
     * @param conn
     * @param statement
     * @param resultSet
     */
    private void close(DruidPooledConnection druidConn, Connection conn, PreparedStatement statement, ResultSet resultSet) {
        try {
            if (resultSet != null)
                resultSet.close();
            if (statement != null)
                statement.close();
            if (conn != null)
                conn.close();
            if (druidConn != null)
                druidConn.close();
        } catch (SQLException e) {
            log.error("close datasource resource hit error!", e);
        }
    }

    /**
     * 给数据源的map中赋值
     *
     * @param dsMap
     * @param newDs
     * @param properties
     */
    public static void putDataSource(Map<String, DruidDataSource> dsMap, DatasourceModel newDs, ZkProperties properties) {
        if (dsMap != null) {
            if (dsMap.isEmpty() || !dsMap.containsKey(newDs.getUrl())) {
                DruidDataSource dataSource = DynamicDataSource.buildDatasource(newDs, properties);
                dsMap.put(newDs.getUrl(), dataSource);
            } else {
                if (!compareDs(dsMap.get(newDs.getUrl()), newDs, properties)) {
                    DruidDataSource newItem = DynamicDataSource.buildDatasource(newDs, properties);
                    dsMap.get(newDs.getUrl()).close();
                    dsMap.remove(newDs.getUrl());
                    dsMap.put(newDs.getUrl(), newItem);
                    log.info("=======CAS new ds, close and remove old ds:{}========", newDs.getUrl());
                }
            }
        }
    }

    /**
     * 比较数据源是否一致
     * @param oldDs
     * @param newDs
     * @param properties
     * @return
     */
    public static boolean compareDs(DruidDataSource oldDs, DatasourceModel newDs, ZkProperties properties) {
        if (oldDs == null && newDs != null)
            return false;
        if (oldDs != null && newDs != null) {
            if (!oldDs.getUsername().equals(newDs.getUsername()))
                return false;
            if (!oldDs.getPassword().equals(newDs.getPwd()))
                return false;
            String url = DynamicDataSource.makeDsUrl(newDs, properties);
            if (!oldDs.getUrl().equals(url))
                return false;
        }
        return true;
    }

    /**
     * 组装查询条件
     * @param query
     * @param condition
     * @param order
     * @param appNameSet
     * @return
     */
    private static String createDbQuery(String query, String condition, String order, Set<String> appNameSet) {
        StringBuilder queryBuilder = new StringBuilder(query);
        if (!appNameSet.isEmpty()) {
            queryBuilder.append(condition);
            for (int i = 0; i < appNameSet.size(); i++) {
                queryBuilder.append(" ?");
                if (i != appNameSet.size() - 1)
                    queryBuilder.append(",");
            }
            queryBuilder.append(")");
        }
        queryBuilder.append(order);

        return queryBuilder.toString();
    }

    /**
     * 对JDBC url进行增强
     * @param url
     * @return
     */
    public static String makeJdbcUrl(String url) {
        if (StringUtils.isNotBlank(url)) {
            if (!url.contains(SOCKET)) {
                String s = "";
                if (url.contains("?")) {
                    s = "&";
                } else {
                    s = "?";
                }
                url += s + SOCKET + TIMEOUT;
            }
        }
        return url;
    }

    /**
     * 初始化routingKeys
     * @param routingKeys
     */
    public void initRouting(Map<String, RoutingKey2DB> routingKeys) {
        log.info("initRouting start");
        DruidPooledConnection druidConn = null;
        Connection conn = null;
        PreparedStatement preparedStatement = null;
        ResultSet resultSet = null;

        int index = 1;
        //从数据库中查询路由关系
        Set<String> appNameSet = properties.getAppName();
        String query = "SELECT tbl_sys_db.dbID as dbID, tbl_sys_db.appName as dbAppName, tbl_sys_db_route.routeKey as routeKey "
                + "FROM tbl_sys_db JOIN tbl_sys_db_route ON tbl_sys_db.dbID=tbl_sys_db_route.dbID "
                + "AND tbl_sys_db_route.action != 2";
        String condition = " AND tbl_sys_db.appName in (";
        String order = " ORDER BY tbl_sys_db.dbID, tbl_sys_db_route.routeKey;";
        String queryCombine = createDbQuery(query, condition, order, appNameSet);
        try {
            druidConn = dataSource.getConnection();
            conn = druidConn.getConnection();
            preparedStatement = conn.prepareStatement(queryCombine);
            for (String appName : appNameSet) {
                preparedStatement.setString(index, appName);
                index++;
            }
            resultSet = preparedStatement.executeQuery();
            RoutingKey2DB routingKey2DB;
            while (resultSet.next()) {
                String dbID = resultSet.getString("dbID");
                String routeKey = resultSet.getString("routeKey");
                String dbAppName = resultSet.getString("dbAppName");

                //给routingKeys中赋值
                DbUrl dbUrl = db_ID.get(dbID);
                routingKey2DB = new RoutingKey2DB(dbUrl.getUrl(), dbUrl.getDbName());
                routingKeys.put(dbAppName + "-tenantID-" + routeKey, routingKey2DB);
            }
            log.info("initRouting finished successfully");
        } catch (SQLException e) {
            log.error("read routing key hit error!", e);
        } finally {
            close(druidConn, conn, preparedStatement, resultSet);
        }
    }

    /**
     * 启动的时候初始化数据库连接
     */
    @PostConstruct
    public void init() {
        dataSource = new DruidDataSource();
        dataSource.setDriverClassName(Constant.DRIVER_CLASS_NAME);
        dataSource.setUsername(properties.getAdminUser());
        dataSource.setPassword(properties.getAdminPwd());
        dataSource.setUrl(makeJdbcUrl(properties.getAdminDB()));
        dataSource.setInitialSize(2);
        dataSource.setMinIdle(1);
        dataSource.setMaxActive(5);
        dataSource.setMaxWait(120000);
        dataSource.setValidationQuery("SELECT 1");
        // 设置检测空闲最大时长
        dataSource.setTimeBetweenEvictionRunsMillis(60000);
        // 空闲时间超过上值时长,测试该连接是否可用
        dataSource.setTestWhileIdle(true);
        // 长时间空闲连接回收
        dataSource.setRemoveAbandoned(true);
        // 设置超过5分钟未使用的连接强制关闭
        dataSource.setRemoveAbandonedTimeout(300);
        // 在取出连接时检查connection的可用性，可保证连接的绝对可用性，如不可用则换其他connection，但是较影响性能
        dataSource.setTestOnBorrow(true);
    }
}

@Data
class DbUrl {
    private String url;
    private String dbName;
}
