package com.lnb.multidatasourcedemo.datasource;

import com.alibaba.druid.pool.DruidDataSource;
import com.google.common.collect.Lists;
import com.lnb.multidatasourcedemo.constant.Constant;

import com.lnb.multidatasourcedemo.curator.CuratorClient;
import com.lnb.multidatasourcedemo.curator.ZkProperties;
import com.lnb.multidatasourcedemo.exception.MultiDatasourceException;
import com.lnb.multidatasourcedemo.holder.DataSourceContextHolder;
import com.lnb.multidatasourcedemo.model.DatasourceModel;
import com.lnb.multidatasourcedemo.model.RoutingKey2DB;
import lombok.Data;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.context.annotation.Primary;
import org.springframework.jdbc.datasource.AbstractDataSource;
import org.springframework.stereotype.Component;

import javax.annotation.PostConstruct;
import javax.sql.DataSource;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;

@Slf4j
@Primary
@Component
public class DynamicDataSource extends AbstractDataSource {

    @Autowired
    private DbAdapter dbAdapter;
    @Autowired
    private CuratorClient curatorClient;
    @Autowired
    private ZkProperties zkProperties;

    private static Map<String, RoutingKey2DB> routingKeys = new ConcurrentHashMap<>();
    private static Map<String, DruidDataSource> dataSources = new ConcurrentHashMap<>();
    private static Object monitorDataSource = new Object();

    /**
     * 启动时路由加载主流程
     */
    @PostConstruct
    public void initDynamicDataSource() {
        try {
            // 初始化数据源，结构是：{url, 数据库连接信息}
            dbAdapter.initDatasource(dataSources);
            //初始化routingkeys，结构是：{租户ID， url}
            dbAdapter.initRouting(routingKeys);
            // 通过读ZK节点添加监听器
            initListener(true);
            log.info("init Dynamic DataSource done!");
        } catch (Exception e) {
            log.error("Init zookeeper error!", e);
            throw new IllegalStateException("Init zookeeper error!");
        }
    }

    /**
     * 通过读ZK节点添加监听器
     */
    public void initListener(boolean isInit) throws Exception {
        log.info("init all zookeeper node Listener start!");
        //监听
        curatorClient.registerAppListenter(this);

        List<AppEntity> allAppDbLists = Lists.newArrayList();
        //获取/dynamicDataSource/dbconfig节点下的所有子节点
        List<String> apps = curatorClient.getChildren(Constant.DBCONFIG_ROOT);
        for (String appName : apps) {
            if (checkAppName(appName, zkProperties.getAppName())) {
                // 获取所有数据库实例
                List<String> dbs = curatorClient.getDBconfig(appName);
                for (String db : dbs) {
                    AppEntity appEntity = new AppEntity(appName, db);
                    allAppDbLists.add(appEntity);
                }
                curatorClient.registerDbconfigListenter(this, appName);
            }
        }
        log.info("register RoutingKey Listener start");
        int allAppDbListSize = allAppDbLists.size();
        CountDownLatch countDownLatch = new CountDownLatch(allAppDbListSize);
        for (AppEntity appDb : allAppDbLists) {
            Thread thread = new Thread(new Participant(curatorClient, this, appDb.getDbName(), appDb.getAppName(), countDownLatch));
            thread.start();
        }
        countDownLatch.await();
        log.info("register RoutingKey Listener finished successfully");
        if (isInit) {
            curatorClient.registerZKstateListenter(this);
        }
        log.info("init all zookeeper node Listener finished successfully!");
    }

    /**
     * 只订阅properties配置的appName,不配则订阅全部
     *
     * @param name zk上DBCONFIG_ROOT下的appName
     * @return
     */
    public static boolean checkAppName(String name, Set<String> setting) {
        boolean r = true;
        if (StringUtils.isBlank(name)) {
            r = false;
        } else {
            if (setting != null && !setting.isEmpty() && !setting.contains(name)) {
                r = false;
            }
        }
        return r;
    }


    /**
     * 最终执行查询时连接数据库
     * @return
     * @throws SQLException
     */
    @Override
    public Connection getConnection() throws SQLException {
        DataSource dataSource;
        String dbName;
        log.debug("start to get connection.......");
        String dataSourceType = DataSourceContextHolder.getDataSourceType();
        log.debug("Current datasource type is [" + dataSourceType + "]");
        if (StringUtils.isBlank(dataSourceType)) {
            throw new MultiDatasourceException("Please set datasource type first!");
        }
        if (!routingKeys.containsKey(dataSourceType)) {
            throw new MultiDatasourceException("Datasource type is error! key:" + dataSourceType);
        }
        RoutingKey2DB key = routingKeys.get(dataSourceType);
        dbName = key.getDatabaseName();
        dataSource = dataSources.get(key.getUrl());
        if (dataSource == null)
            throw new MultiDatasourceException("Local key " + dataSourceType + " not exist!");
        Connection connection = dataSource.getConnection();
        // 切换数据库实例下实际DB名称
        connection.setCatalog(dbName);
        log.debug("end to get connection!");
        return connection;
    }

    @Override
    public Connection getConnection(String username, String password) throws SQLException {
        throw new UnsupportedOperationException("Could not get datasource by username!");
    }

    /**
     * 创建一个数据源
     * @param dataSourceModel
     * @param properties
     * @return
     */
    public static DruidDataSource buildDatasource(DatasourceModel dataSourceModel, ZkProperties properties) {
        DruidDataSource ds = new DruidDataSource();
        if (!dataSourceModel.validateDatasourceModel()) {
            log.error("dataSourceModel is not valid:{}", dataSourceModel);
            throw new MultiDatasourceException("Datasource config has empty value! Please check again.");
        }

        // 取设置的默认值
        ds.setDriverClassName(Constant.DRIVER_CLASS_NAME);
        ds.setUsername(dataSourceModel.getUsername());
        //根据配置判断连接从库还是主库，得到最终的连接url
        String url = makeDsUrl(dataSourceModel, properties);
        log.info("build datasource ! url:{}", url);
        ds.setUrl(url);
        ds.setPassword(dataSourceModel.getPwd());
        ds.setInitialSize(
                dataSourceModel.getInitialSize() != null ? Integer.valueOf(dataSourceModel.getInitialSize()) : 5);
        ds.setMaxActive(dataSourceModel.getMaxActive() != null ? Integer.valueOf(dataSourceModel.getMaxActive()) : 80);
        ds.setMinIdle(dataSourceModel.getMinIdle() != null ? Integer.valueOf(dataSourceModel.getMinIdle()) : 10);
        // 配置从连接池获取连接等待超时的时间
        ds.setMaxWait(dataSourceModel.getMaxWait() != null ? Integer.valueOf(dataSourceModel.getMaxWait()) : 10000);
        ds.setValidationQuery(
                (dataSourceModel.getValidationQuery() != null && !"".equals(dataSourceModel.getValidationQuery()))
                        ? dataSourceModel.getValidationQuery() : "SELECT 1");
        ds.setValidationQueryTimeout(1000);
        // 配置一个连接在池中最大空闲时间，单位是毫秒
        ds.setMinEvictableIdleTimeMillis(60000);
        // 设置检测空闲最大时长
        // 配置间隔多久启动一次 DestroyThread，对连接池内的连接才进行一次检测，单位是毫秒。
        // 检测时:
        // 1.如果连接空闲并且超过minIdle以外的连接，如果空闲时间超过minEvictableIdleTimeMillis设置的值则直接物理关闭。
        // 2.在minIdle以内的不处理。
        ds.setTimeBetweenEvictionRunsMillis(60000);
        // 空闲时间超过上值时长,测试该连接是否可用
        // 设置从连接池获取连接时是否检查连接有效性，true 时，如果连接空闲时间超过 minEvictableIdleTimeMillis 进行检查，否则不检查; false时，不检查
        ds.setTestWhileIdle(true);
        // 长时间空闲连接回收
        ds.setRemoveAbandoned(true);
        // 设置超过多长时间未使用的连接强制关闭,秒
        ds.setRemoveAbandonedTimeout(150);
        // 关闭abanded连接时输出错误日志，这样出现连接泄露时可以通过错误日志定位忘记关闭连接的位置
        ds.setLogAbandoned(true);
        // 在取出连接时检查connection的可用性，可保证连接的绝对可用性，如不可用则换其他connection，但是较影响性能
        ds.setTestOnBorrow(true);
        // 设置往连接池归还连接时是否检查连接有效性，true 时，每次都检查；false 时，不检查
        ds.setTestOnReturn(false);
        // 打开后，增强timeBetweenEvictionRunsMillis的周期性连接检查，minIdle内的空闲连接，每次检查强制验证连接有效性.
        // 参考：https://github.com/alibaba/druid/wiki/KeepAlive_cn
//        ds.setKeepAlive(true);
        return ds;
    }

    /**
     * 主从切换
     * @param dataSourceModel
     * @param properties
     * @return
     */
    public static String makeDsUrl(DatasourceModel dataSourceModel, ZkProperties properties) {
        String url;
        if (!StringUtils.isBlank(properties.getReadOnly()) && "true".equalsIgnoreCase(properties.getReadOnly()))
            url = dataSourceModel.getUrl().replace(Constant.WARITE,
                    Constant.READ + (StringUtils.isBlank(properties.getSlot()) ? "" : properties.getSlot()));
        else
            url = dataSourceModel.getUrl();

        url = DbAdapter.makeJdbcUrl(dataSourceModel.getUrlPrefix() + url + dataSourceModel.getCommonDB());

        return url;
    }

    static class Participant implements Runnable {
        private DynamicDataSource dynamicDataSource;
        private String dbName;
        private String appName;
        private CuratorClient curatorClient;
        private CountDownLatch countDownLatch;

        public Participant(CuratorClient curatorClient, DynamicDataSource dynamicDataSource, String dbName, String appName, CountDownLatch countDownLatch) {
            this.dynamicDataSource = dynamicDataSource;
            this.curatorClient = curatorClient;
            this.dbName = dbName;
            this.appName = appName;
            this.countDownLatch = countDownLatch;
        }

        public void run() {
            try {
                curatorClient.registerRoutingKeyListenter(dynamicDataSource, dbName, appName);
            } catch (Exception e) {
                e.printStackTrace();
            } finally {
                countDownLatch.countDown();
            }
        }
    }

    /**
     * 监听到的数据添加到routingkeys中
     * @param dbUrl
     * @param routingKeyList
     * @param appName
     * @throws Exception
     */
    public void registerRoutingKey(String dbUrl, List<String> routingKeyList, String appName) throws Exception {
        for (String routingKey : routingKeyList) {
            String dbName = curatorClient.readWithPath(Constant.getAppDB(appName), dbUrl, routingKey);
            if (StringUtils.isBlank(dbName))
                return;
            RoutingKey2DB routingKey2DB = new RoutingKey2DB(dbUrl, dbName);
            synchronized (monitorDataSource) {
                routingKeys.put(routingKey, routingKey2DB);
            }
            log.debug("Add routingKey [" + routingKey + "] DB [" + routingKey2DB + "]");
        }
    }

    /**
     * 从routingkeys中删除数据
     * @param routingKey
     */
    public void removeRoutingKey(String routingKey) {
        if (!routingKeys.containsKey(routingKey))
            return;
        synchronized (monitorDataSource) {
            Iterator<Map.Entry<String, RoutingKey2DB>> rs = routingKeys.entrySet().iterator();
            while (rs.hasNext()) {
                Map.Entry<String, RoutingKey2DB> next = rs.next();
                if (routingKey.equalsIgnoreCase(next.getKey())) {
                    logger.info("Remove routing key : [" + next.getKey() + "]");
                    rs.remove();
                }
            }
        }
    }

    /**
     * 往dataSources中添加数据
     * @param dataSourceModel
     */
    public void registerDataSource(DatasourceModel dataSourceModel) {
        synchronized (monitorDataSource) {
            String key = dataSourceModel.getUrl();
            if (dataSources.containsKey(key)) {
                logger.info("dataSourceKey [" + key + "] have existed ");
                removeDatasource(key);
            }
            DruidDataSource ds = buildDatasource(dataSourceModel, zkProperties);
            dataSources.put(key, ds);
            logger.info("put [" + key + "] datasource [" + ds + "]");
        }
    }

    /**
     * 从datasource中删除数据
     * @param key
     */
    public void removeDatasource(String key) {
        if (!dataSources.containsKey(key))
            return;
        if (zkProperties.getAppName() == null && zkProperties.getAppName().isEmpty()) {
            log.info("listen all appName, don't remove datasource!");
            return;
        }
        synchronized (monitorDataSource) {
            Iterator<Map.Entry<String, DruidDataSource>> ds = dataSources.entrySet().iterator();
            while (ds.hasNext()) {
                Map.Entry<String, DruidDataSource> next = ds.next();
                if (key.equalsIgnoreCase(next.getKey())) {
                    logger.info("Remove datasource key : [" + next.getKey() + "]");
                    next.getValue().close();
                    ds.remove();
                    break;
                }
            }
        }
    }
}

@Data
class AppEntity {

    public AppEntity(String appName, String dbName) {
        this.appName = appName;
        this.dbName = dbName;
    }

    private String appName;
    private String dbName;
}


